/*++++++++++++++
.IDENTIFICATION catstd.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS
.VERSION  1.0   15-May-1991 (anafile)
.VERSION  4.0   28-Sep-1994 New package
.VERSION  4.1   09-Nov-1994: Add equinox + Epoch in TABINFO
.VERSION  4.2   20-Nov-1995: Allow edition for HTML
.VERSION  4.3   08-Jan-1996: Corrected Bug edition with consecutive Byte-by-Byte
.VERSION  4.4   21-Feb-1996: New keywords, Objects
.VERSION  4.5   16-Mar-1996: Accept GCPD/photometry
.VERSION  4.6   26-Apr-1996: P_opt is a number, several formats.
.VERSION  4.7   04-Apr-1997: Added structure signatures
.VERSION  4.8   30-May-1997: Update the TeX output
.VERSION  4.9   02-Sep-1997: Set FK for r_ columns
.VERSION  4.91  07-Apr-1998: Added the secno number
.VERSION  4.92  23-Dec-1998: Verify Objects 
.VERSION  4.93  07-May-1999: Accept ReadMe from stdin
.VERSION  5.0   16-Feb-2000: More checks on titles
.VERSION  5.1   09-Dec-2000: cad_name (name of catalog)
.VERSION  5.2   19-Apr-2001; Add the credate
.VERSION  5.3   13-Jul-2001: Be sure to remove the "rule" in File Summary
.VERSION  5.4   01-Apr-2002: Accept TABMAP
.VERSION  5.5   03-Nov-2002: cad_tdup
.VERSION  5.6   06-Jan-2003: Accept a directory name (append 'ReadMe').
.VERSION  5.7   27-Feb-2003: Accept a range of years
.VERSION  5.71  11-Apr-2004: F_Aunit in FILINFO indicates "tabform" allocated.
.VERSION  5.72  17-Apr-2004: \3PAR for a PARAMETER
.VERSION  5.73  13-May-2004: bug in check_keyline (when several lines)
.VERSION  5.74  30-Dec-2004: Assignemnt of FK/PK in case of many files
			sharing the same description: speedup
.VERSION  5.75  11-Mar-2005: File is "readme.viz" if exists.
.VERSION  5.8   20-Jul-2005: Data file may contain headers
.VERSION  5.81  26-Apr-2006: file->reclen is adjusted to largest record.
.VERSION  5.82  02-Jun-2007: Cosmetics (-Wall)
.VERSION  5.82  25-Nov-2007: cad_oid()
.VERSION  5.83  28-Nov-2007: Other journals.
.VERSION  5.84  02-Jul-2008: Take care of repetition
.VERSION  5.85  17-Jan-2009: Pb with table without columns...
.VERSION  5.86  09-Oct-2009: Links have a title (use of yAref)
.VERSION  5.87  14-Oct-2009: SEC_History
.VERSION  5.88  16-Oct-2009: No error if several authors
.VERSION  5.89  15-Dec-2009: 64-bit
.VERSION  5.90  20-Jan-2010: Addition SEC_Abstract + SEC_Ack
.VERSION  5.91  14-Aug-2010: Verify all ascii chars
.VERSION  5.92  11-Jun-2013: cad_print: print also the headlines
.VERSION  5.93  23-Jul-2013: bug!!
.COMMENTS       Get COLINFO from Intro/ReadMe file
		The FTP_Cat environment variable is asssumed for
		catalogue location via FTP_Cat
---------------*/

#ifndef DEBUG
#define DEBUG 0
#endif
#ifndef TEST
#define TEST 0
#endif
#ifndef FETCH			/* Generate Fetch Files Boxes	*/
#define FETCH 0
#endif
/*static int vartest = TEST; */

#include <adckeys.h>		/* ADC Keywords			*/
#include <journals.h>		/* Known journal abbreviations 	*/
#include <catstd.h>
#include <filutil.h>		/* bzero, etc */
#include <strings.h>		/* bzero, etc */
#include <time.h>		/* time() fct */
#include <fcntl.h>		/* File Control	   */
#include <sys/stat.h>		/* Parameters of Files	*/
#include <sys/types.h>		/* Parameters of Files	*/
#ifndef O_BINARY
#define O_BINARY 0
#endif

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])
#define iskeysep(c)	(((c)==';') || ((c)=='-'))
#define starts_by(s,t)  (!strncmp(s,t,sizeof(t)-1))

int fputc();

static char catid[80];	/* Catalogue designation */
static char *Tbf, *Trm;	/* \\bf and \\rm flags	 */
int (*FPUTC)()  	/* Function which writes a character */
	= fputc ;
static char Tget[2];	/* (A) (F) Anchors	 */
static char rule_char = '-';	/* Rule Char	 */

static CATINFO *catinfo ;	/* For cad_print */
typedef struct cbuf {
    struct cbuf *next ;
    int used ;
    char buf[8<<10] ;		/* 8K */
} CBUF ;			/* Chained Buf.  */
static CBUF *loaded_buf ;


#if DEBUG || TEST
static char inbyte;
#endif

/* Signatures of the structures */
static char *signatures[] = {
    /* 0 */ "\3COL", 
    /* 1 */ "\3TAB", 
    /* 2 */ "\3FIL", 
    /* 3 */ "\3CAT", 
    /* 4 */ "\3SEC", 
    /* 5 */ "\3MAP", 
    /* 6 */ "\3PAR", 	/* PARAM -- added v5.72 */
} ;
static void *free_oid;

static char *aanchor = "http://vizier.u-strasbg.fr" ;

/*==================================================================
		Internal Utilities
 *==================================================================*/

static long load_stdin()
/*++++++++++++++++
.PURPOSE  Load the stdin into a set of chained buffers.
.RETURNS  Number of bytes
.REMARKS  loaded_buf is set to the beginning.
-----------------*/
{
  long bytes = 0 ;
  CBUF *buf0 ; 
  char *b, *e ;
  int len ;

    buf0 = loaded_buf = (CBUF *)malloc(sizeof(CBUF)) ;
    b = buf0->buf ; e = b + sizeof(buf0->buf) ;
    while(1) {
	if (b >= e) {
	    buf0->used = b - buf0->buf ;
	    buf0->next = (CBUF *)malloc(sizeof(CBUF)) ;
	    buf0 =  buf0->next ;
            b = buf0->buf ; e = b + sizeof(buf0->buf) ;
	}
	len = read(0, b, e-b) ;
	if (len <= 0) break ;
	bytes += len ;
	b += len ;
    }
    buf0->used = b - buf0->buf ;
    buf0->next = (CBUF *)0 ;
    return(bytes) ;
}

static long load_move(target)
/*++++++++++++++++
.PURPOSE  Move the loaded_buf
.RETURNS  Number of bytes
.REMARKS  loaded_buf is freed.
-----------------*/
  char *target ;	/* OUT: Where to move loaded_buf */
{
  CBUF *b, *n ;
  char *t ;
    t=target, b=loaded_buf; 
    while(b) {
	memcpy(t, b->buf, b->used) ;
	t += b->used ;
        n = b->next ;
	free(b) ;
        b = n ;
    }
    return(t-target) ;
}

#if 0
static int tex_fputc(c, fh)
/*++++++++++++++++
.PURPOSE  Issue a character in TeX mode (table care of special's)
.RETURNS  0
-----------------*/
  int c;	/* IN: The character to issue 	*/
  FILE *fh;	/* IN: File output		*/
{
  static char escape[128] ;
    if (! escape[0]) {	/* Fill the table of special characters */
	escape['{'] = escape['}'] = escape['%'] = escape['&'] = '\\';
	escape[0] = 1 ;
    }
    if(escape[c&0177]) fputc(escape[c&0177], fh);
    fputc(c, fh);
    return(0);
}
#endif

static int sec_class(header)
/*++++++++++++++++
.PURPOSE  Find the class (SEC_xx) of a header
.RETURNS  The corresponding numeric
-----------------*/
  char *header;     /* IN: Text to analyze */
{
    if (stumatch("Byte-*-byte Description*:*",header))
    	return(SEC_BytebyByte);
    if (strmatch("File Summary:", header))
    	return(SEC_FileSummary);
    if (strmatch("Note (*):*", header)) 
    	return(SEC_Note);
    if (strmatch("Note on*:*", header)) 
    	return(SEC_Note);
    if (strmatch("Notes on*:*", header)) 
    	return(SEC_Note);
    if (strmatch("Description*", header)) 
    	return(SEC_Description);
    if (strmatch("Abstract[ :]*", header)) 
    	return(SEC_Abstract);
    if (strmatch("Objects[ :]*", header)) 
	return(SEC_Objects);
    if (strmatch("Object:", header)) 
	return(SEC_Objects);
    if (stumatch("See also:",header))
	return(SEC_SeeAlso);
    if (strmatch("References*", header)) 
    	return(SEC_References);
    if (strmatch("(End)*", header)) 
    	return(SEC_End);
    if (strmatch("Keywords:*", header)) 
    	return(SEC_Keywords);
    if (strmatch("Mission_Nam*:*", header)) 
    	return(SEC_Keywords);
    if (strmatch("Acknowledgements[ :]*", header)) 
    	return(SEC_Ack);
    if (strmatch("History*:*", header)) 
    	return(SEC_History);
    if (strmatch("Nomenclature *", header)) 
    	return(SEC_Naming);
    if (strmatch("*_Keywords:*", header)) 
    	return(SEC_Keywords);
    if (strmatch("Related Catalo*:*",header))
	return(SEC_SeeAlso);
    if (stumatch("Related Data*:*",header))
	return(SEC_SeeAlso);
    if (strmatch("Introduction*", header)) 
    	return(SEC_Description);
    /* For Future Use: */
    if (strmatch("Column Description*:*", header))
    	return(SEC_Col);
    return(SEC_Unknown);
}

static int print_cgi(text, fout)
/*++++++++++++++++
.PURPOSE  Write the text, including links to files / sections
.RETURNS  Number of links generated
-----------------*/
  char *text;     /* IN: Text to print */
  FILE   *fout;	  /* IN: Opened output file 	*/
{
  FILINFO *pf ;
  SECTION *ps ;
  char *t, *p;
  int i, x, nlinks;

    nlinks = 0 ;
    for (t=text; *t; t++) {
	(*FPUTC)(*t, fout); 
	if (*t != '"') continue ;

	/* Here, text starting by the Quote " */
	t++ ;
	i = strloc(t, '"') ;
	if (!t[i]) {	/* non-matched quote */
	    (*FPUTC)(*t, fout);
	    continue ;
	}
	x = t[i]; t[i] = 0 ;

	/* Is it a file name ? If yes, generate link -- if not too large */
	if ((pf = cad_findfile(catinfo, t))) {
	    if (pf->records > 1000) fprintf(fout, "%s%s%s", Tbf, t, Trm) ;
	    else fprintf(fout, "\\glutag{getCatFile %s/%s|%s}", catid, t, t) ;
	    t[i] = x ;
	    t += i;
	    fputc(*t, fout);	/* Write Closing Quote */
	    nlinks++ ;
	    continue ;
	}

	/* Is it a section ? If yes, write the link... */
	for (p = t+i+1 ; isspace(*p); p++) ;
	if (stumatch("section*", p)) {
	    if ((ps = sec_find(catinfo->section[0], t))) 
	        fprintf(fout, "\\Aref{%s}{%s}\"", sec_id(ps), t) ;
	    else fprintf(fout, "%s\"{\\fg{red}(?)}", t) ;
	    t[i] = x ;
	    t += i;		/* Quote edited */
	    nlinks++ ;
	    continue ;
	}
	(*FPUTC)(*t, fout);
	t[i] = x ;
    }

    return(nlinks) ;
}

static int isanchor(text)
/*++++++++++++++++
.PURPOSE  Check whether the first word of text is an anchor
.RETURNS  0/1
-----------------*/
  char *text;     /* IN: Text to analyze */
{
  char *p;
    for (p=text; (*p == 'I') || (*p == 'V') || (*p == 'J'); p++ ) ;
    if ((*p == '/') && (p != text)) return (1);
    if (strmatch("GCPD/*", text)) return(1);
    if (stumatch("http:*", text)) return(1);
    if (stumatch("ftp:*", text)) return(1);
    return(0);
}

static char *jfind(abbr)
/*++++++++++++++++
.PURPOSE  Find the full journal
.RETURNS  The full name
-----------------*/
  char *abbr;     /* IN: Abbreviation of the Journal */
{
  int i; char *a, *p, **j;
    if (starts_by(abbr, "J/")) abbr += 2;
    for (i=ITEMS(journal)/2, j=journal; --i>=0; j += 2) {
	for (a=abbr, p=*j; *p == *a ; p++, a++) ;
	if (*p) continue ;
	if (*a == '/') return(j[1]) ;
    }
    return ((char *)0);
}

static int str_text(str, len)
/*++++++++++++++++
.PURPOSE  Find out the end of the text: remove ending spaces,
	and ending rules
.RETURNS  the length
-----------------*/
  char *str;	/* IN: String to examine */
  int  len;	/* IN: Length of string  */
{
  char *p, *e; int i;
#if DEBUG>1
    printf("....str_text of text of %d bytes:\n", len);
    fwrite(str, 1, len, stdout); 
#endif
    for (i = len; i > 0; i = (p-str)) {
#if DEBUG>1
	printf("....i=%d", i);
#endif
    	i = strnline(str, i);	/* Remove ending spaces */
#if DEBUG>1
	printf("=>%d\n", i);
#endif
    	if (i <= 20) break;
    	for (p = e = str-1+i; (*p == *e) && (p > str); p--) ;
	if (p == str) continue;	/* Only a rule, no data	*/
    	if ((*p == '\n') && match_rule(p+1)) continue;
        break;
    }
#if DEBUG>1
    printf("...now of length %d", i); inbyte = getchar();
#endif
    if (i<0) return(0);
    return(i);
}

static int strword(str, word)
/*++++++++++++++++
.PURPOSE  Check whether "word" is a word in str
.RETURNS  0 (not in ) / > 0 (word is in str)
-----------------*/
  char *str ;	/* IN: The string to scan	*/
  char *word ;	/* IN: The word to look for	*/
{
  int i, bw ; char *p; 
    i = strlen(word) ;
    for (bw=1, p=str; *p; p++) {
	if (!bw) { if (!isalnum(*p)) bw=1; continue ; }
	bw = !isalnum(*p) ;
	if (*p != *word) continue ;
	if (strncmp(p, word, i) != 0) continue ;
	if (!isalnum(p[i])) return(1) ;
    }
    return(0) ;
}

static int check_file(pf, fn)
/*++++++++++++++++
.PURPOSE  Verify the consistency of file characteristics
.RETURNS  Number of errors + warnings
-----------------*/
  FILINFO *pf;	/* IN: The file parameters	*/
  char *fn;	/* IN: Name of "ReadMe" file */
{
  static char acceptable_ext[] = "fits html";
  int nw, lineno, i; char *p, *x, *err;
  
    nw = 0; lineno = pf->lineno+1;
    if (strdiff(fn, pf->dfile) == 0) {	/* The Description File */
    	if (pf->reclen != 80) nw++, fprintf(stderr, 
    	    "++++ line %d: Lrecl of file '%s' should be 80\n", 
    	    lineno, pf->dfile);
    	if (pf->records != 0) nw++, fprintf(stderr, 
    	    "++++ line %d: Records of file '%s' should be unspecified (.)\n", 
    	    lineno, pf->dfile);
    	return(nw);
    }

    /* Verify MS-DOS compatibility */
    for (i=ITEMS(mixed_names); --i >= 0; ) {
	if (strmatch(mixed_names[i], pf->dfile)) return(nw);
    }

    err = (char *)0;
    i = strloc(pf->dfile, '.'); 
    x = pf->dfile+i; if (*x) x++;	/* Extension */
    /* Verify the file has a description */
    if ((!pf->tabform) && (strcmp(x, "dat") == 0) 
       && ((pf->flags&F_NODB) == 0)) fprintf(stderr,
    	"++++ line %d: file not described: %s\n", lineno, pf->dfile) ;
    if (i>8) { if (!strchr(pf->dfile, '/')) err = "length > 8"; }
    else if (((int)strlen(x)>3) && (!strstr(acceptable_ext, x))) 
	err = "extension > 3char";
    for (p=pf->dfile, i=0; *p && (!err); p++) {
    	if (islower(*p)) continue;
    	if (isdigit(*p)) continue;
    	if (*p == '_' || *p == '-') continue;
	if ((*p == '/') /*&& (pf->reclen == 0)*/) {
	    if ((i==0) && ((p[1]=='*')||(!p[1]))) 
		return(nw);				/* DIRECTORY */
	}
    	if (*p == '.') {
    	    if (i) err="Too many dots";
    	    i++;
    	    continue;
        }
    	if (isupper(*p)) err = "don't use upper-case letters in filenames"; 
    	else err = "only digits, letters, _ and - allowed";
    }
    if (err) nw++, fprintf(stderr, 
    	"++++ line %d: bad filename '%s' (%s)\n",
    	lineno, pf->dfile, err);
    	
    	/* Verify lrecl, number of records */
    if (*x == 'f') {
    	if (strmatch("fit*", x) && (pf->reclen != 2880)) nw++, fprintf(stderr,
    	    "++++ line %d: FITS file '%s' should have Lrecl=2880 (is %d)\n", 
    	    lineno, pf->dfile, pf->reclen);
    	if (strmatch("fih", x) && (pf->reclen > 80)) nw++, fprintf(stderr,
    	    "++++ line %d: FITS header '%s' should have Lrecl<=80 (is %d)\n", 
    	    lineno, pf->dfile, pf->reclen);
    }
    if ((pf->reclen<8)&&(pf->reclen>0)) nw++, fprintf(stderr, 
    	"++++ line %d: '%s' has a Lrecl of %d??\n",
    	lineno, pf->dfile, pf->reclen);
    if (pf->records < 1) nw++, fprintf(stderr, 
    	"++++ line %d: '%s' has %d records??\n",
    	lineno, pf->dfile, pf->records);

    return(nw);
}

static int set_catid(ps, fn)
/*++++++++++++++++
.PURPOSE  Copy into the catid static, the catalogue identification
.RETURNS  Number of errors
-----------------*/
  SECTION *ps;	/* IN: The section to check 	*/
  char *fn;	/* IN: Name of "ReadMe" file 	*/
{
  int nw, i; char *p;
    nw = 0; catid[0] = 0;
    if (ps->class != SEC_Title) {
    	nw++, fprintf(stderr, "++++Title missing in '%s'++++\n", 
	    fn ? fn : "(stdin)");
    	return(nw);
    }
				/* Get catalogue designation */
    for (p = ps->abody; isgraph(*p); p++) ;
    i = p - ps->abody; i = MIN (i, sizeof(catid)-1);
    strncpy(catid, ps->abody, i); catid[i] = 0;
    return(nw);
}

char *cad_name(catinfo)
/*++++++++++++++++
.PURPOSE  Returns the catalog identification (in a static area)
.RETURNS  e.g. J/A+AS/180/123 / ""
-----------------*/
  CATINFO *catinfo ;
{
  SECTION *ps ;
    if (!catinfo) return((char *)0) ;
    ps = catinfo->section[0] ;
    if (ps->class == SEC_Title) set_catid(catinfo, catinfo->ffile) ;
    else return("") ;
    return(catid) ;
}

static int check_keyline(akeys, aline, eline, lineno)
/*++++++++++++++++
.PURPOSE  Verify the consistency of one line of keywords
.RETURNS  Number of errors + warnings
-----------------*/
  char **akeys;	/* IN: The lisdt of accepted keywords */
  char *aline ;	/* IN; The line with keywords */
  char *eline ;	/* IN: End of line address... */
  int lineno ;	/* IN/ Line number in ReadMe file (for errors) */
{
  char **a, *p, *n, *e=(char *)0, x;
  int nw = 0 ;

    for (p=aline; p<eline; p=n) {
	while (isspace(*p) && (p<eline)) {
	    if (*p == '\t') fprintf(stderr, 
	      "++++Warning: there are tab(s) in ReadMe file\n");
	    p++;
	}
	n = p ;
	if (iskeysep(*n)) {	/* Skip separation */
	    while ((n<eline) && (*n != ' ')) n++ ;
	    continue ;
	}
	if (n >= eline) continue;
	/* Keep words until next separator */
	while ((n<eline) && (! iskeysep(*n))) {
	    while (*n && (!isspace(*n))) n++ ;
	    if (n[-1] == ';') --n ;
	    e = n ;	/* End of word */
	    while (isspace(*n)) n++ ;
	}
	x = *e ; *e = 0 ;
	for (a=akeys; *a && strcmp(*a, p); a++) ;
	if (!*a) nw++, fprintf(stderr, 
	    "++++ line %d: keyword not in ADClist: %s\n", lineno, p) ;
	*e = x ;
    }
    return(nw) ;
}

static int check_keywords(ps, fn)
/*++++++++++++++++
.PURPOSE  Verify the consistency of the title header
.RETURNS  Number of errors + warnings
-----------------*/
  SECTION *ps;	/* IN: The section to check */
  char *fn;	/* IN: Name of "ReadMe" file */
{
  int nw, lineno ;
  char *p, *n ;

    lineno = ps->lineno + 1; 
    nw = 0 ;
    
    if ((p = ps->header)) {
	p += strloc(p, ':') ;
	if (*p) p++ ;
	nw += check_keyline(ADCkeys, p, p+strlen(p), lineno) ;
    }

    for (p=ps->abody; p<ps->ebody; p=n+1) {
	lineno++ ;
	n = p + strloc(p, '\n') ;
	nw += check_keyline(ADCkeys, p, n, lineno) ;
    }

    return(nw) ;
}

static int check_see(ps)
/*++++++++++++++++
.PURPOSE  Verify the "See also" section
.RETURNS  Number of errors + warnings
-----------------*/
  SECTION *ps;	/* IN: The section to check */
{
  int nw, lineno, i, T ;	
  /* T is set to 1 when catalogue title MUST exist */
  char *p, *n;

    lineno = ps->lineno + 1; 
    nw = 0 ;
    T  = 1 ;	/* Check in first line */
    
    for (p = ps->abody; p < ps->ebody; p = n+1, T = 0) {
	n = p + strloc(p, '\n') ;
	lineno++ ;
	while (*p == ' ') p++ ;
	for (i=T; isgraph(*p); p++) if(*p == '/') i++;	/* Skip first word */
	while ((p[0] == ' ') && (p[1] == ' ')) p++ ;
	if ((p[0] == ' ') && (p[1] == ':') && isspace(p[2])) continue ;
	if (i>0) nw++, fprintf(stderr, 
	    "++++ line %d: colon-blank ( : ) separator missing in See Also\n",
	    lineno);
    }

    return(nw) ;
}

static int check_title(ps, fn)
/*++++++++++++++++
.PURPOSE  Verify the consistency of the title header
.RETURNS  Number of errors + warnings
-----------------*/
  SECTION *ps;	/* IN: The section to check */
  char *fn;	/* IN: Name of "ReadMe" file */
{
  static char *err[] = {	/* Error messages for phase */
        "Title (left-flushed)",	/* Title not found */
  	"Authors (indented)",
  	"<Reference> (indented)",
  	(char *)0 ,
  	(char *)0
  };
  int nw, phase, lineno, i, year[2], year3=0, year4, nref=0; 
  char *p, *n, *e, *user, x, A=0;
  char reft[80];		/* Reference Template "*Journal*Volume*Page*" */

    lineno = ps->lineno; 
    reft[0] = 0;
    /* From 'nobody' execution, diminish the number of warnings */
    user = getenv("USER");
    if (user && strcmp(user, "nobody") == 0) user = (char *)0;

				/* Get catalogue designation */
    nw = set_catid(ps, fn);
    if (ps->class != SEC_Title) return(nw);
    
    if (ps->header) nw++, lineno++, fprintf(stderr,
    	"++++ line %d: colon (:) not allowed in Title\n", lineno);
    
    if (starts_by(catid, "J/")) {	/* Generate template */
	i = starts_by(catid, "J/other") ? 8 : 2 ;	/* Mod. V5.83 */
	i += strloc(catid+i, '/'); 			/* Mod. V5.83 */
	n = jfind(catid); 
	if (n) {	/* Journal exists, n = full name */
	    reft[0] = '*'; strcpy(reft+1, n); 
	    n = reft + strlen(reft); *(n++) = '*';
	    for (p = catid+i+1; *p; p++)
		*(n++) = (*p == '/')||(*p == '.') ? '*' : *p;
	    *(n++) = '*'; *n = 0;
	}
	else nw++, fprintf(stderr,
	    "++++ line %d: Journal abbreviation is unknown: %s\n", 
	    lineno, catid);
    }
				/* Verify first line */
    i = strloc(ps->abody, '\n'); n = ps->abody + i; lineno++;
    p = ps->abody + strnline(ps->abody, i); 
    x = *p; *p = 0;		/* Terminate string */
    if (! isupper(*(ps->abody))) nw++;
    if (!strmatch("* * (* *)", ps->abody)) {
	nw++;
	if (strmatch("* * (*)", ps->abody)) 
	    e = "no blank between author and year" ;
	else e = "no author, year" ;
	fprintf(stderr, "++++ line %d: %s\n", lineno, e) ;
    }
    if (nw) fprintf(stderr, 
      "++++ line %d: Title should be <Cat-id\ttitle (author, year)>\n\t<%s>\n",
      lineno, ps->abody);
    /* Find out the years of publication */
    for (--p; isgraph(*p); --p) ;
    p++; year[0] = year[1] = atoi(p); while (isdigit(*p)) p++;
    if (*p == '-') {		/* Range of years */
	++p;
	year[1] = atoi(p); 
	if (year[1] < 100) year[1] += 100*(year[0]/100);
	if (year[1] < year[0]) year[1] += 100;
    }
    /*  year = atoi(p-5) ;	-- Last Year of Publication */
    p += strlen(p);
    *p = x; 			/* Reset string	*/

	/* Verify the existence of title / authors / refs */
	/* Phase: 1=Title, 2=Authors, 3=Ref */
    for (phase=0, p=n+1; p<ps->ebody; p=n+1) {
    	lineno++;
    	n = p + strloc(p, '\n');
#if DEBUG
	x = *n; *n = 0; printf("line=%d, phase=%d: %s\n", lineno, phase, p); 
	*n=x;
#endif
    	if (match_rule(p)) continue;
    	if (isgraph(*p)) { 	/* Title (non-indented) */
	    switch(phase) {
	      case 1: continue ;
	      case 2:	/* Title preceded by Authors */
		fprintf(stderr, "++++ line %d: No reference?\n", lineno-1) ;
		nw++;	/* NO BREAK */
	      case 0: case 3: case 4:
		A = 0 ;	/* No Author (yet) */
		phase = 1 ;
		year3 = 0 ;
	    }
    	    continue;
    	}
    	while (*p == ' ') p++;
    	if (isalpha(*p)) {	/* Author */
    	    if (phase == 1) { phase = 2; A = toupper(*p); }
    	    if (phase != 2) nw++, fprintf(stderr, 
    	    	"++++ line %d: Authors misplaced\n", lineno);
    	    continue;
    	}
    	if (*p == '<') {	/* Reference */
    	    if (phase == 2) phase = 3;
    	    if (phase != 3) nw++, fprintf(stderr, 
    	    	"++++ line %d: Reference misplaced\n", lineno);
	    if (reft[0]) {	/* Check against template */
		if (!strmatch(reft, p)) {
		    nw++;
		    fprintf(stderr, "++++ line %d: ", lineno);
		    fwrite(p, 1, strloc(p, '\n'), stderr);
		    fprintf(stderr, " doesn't match\n%13s<%s>\n", "", reft);
		}
	    }
	    i = strloc(p, '>') ;
	    if (!p[i]) { 
		fprintf(stderr, "++++ line %d: reference not closed\n", lineno);
		nw++;
		continue ;
	    }
	    p += i ;	/* p->closing angle bracket */
	    while (n < p) { 	/* Reference on several lines */
		if (*n) n++; else break ;
		n += strloc(n, '\n') ;
		lineno++ ;
	    }
	    while (*--p == ' ') ;
	    if (*p == /*(*/')') {
		p-- ;
		while (isdigit(*p)) --p ;
		year3 = atoi(p+1) ;
		if ((year3<year[0])||(year3>year[1])) nw++, fprintf(stderr,
		    "++++ line %d: year not compatible (%d vs %d-%d)\n", 
		    lineno, year3, year[0], year[1]) ;
	    }
    	    continue;
    	}
    	if (*p == '=') {	/* BibCode */
	    if (phase < 3) nw++, fprintf(stderr,
		"++++ line %d: Bibcode misplaced\n", lineno);
	    phase = 4 ;
	    year4 = atoi(p+1) ;
	    if ((year4<year[0])||(year4>year[1])) nw++, fprintf(stderr, 
		"++++ line %d: year in bibcode not compatible (%d vs %d-%d)\n", 
		lineno, year4, year[0], year[1]) ;
	    /* if (year3 && (year4 != year3)) nw++, fprintf(stderr, 
		"++++ line %d: year in bibcode not compatible (%d vs %d)\n",
		lineno, year4, year3) ; */
	    /* No error message when 'nobody' */
	    /* V5.88: No error message when several bibcodes (only first) */
	    if (A && (p[19] != A) && user && (nref == 0)) {
		nw++;
		x = p[20]; p[20] = 0 ;
		fprintf(stderr, "++++ line %d: incorrect BibCode: %s", 
		    lineno, p+1) ;
		if ((strlen(p) < 20) || isspace(p[19])) fprintf(stderr,
		    " (too short)") ;
		else fprintf(stderr, " (author's initial should be '%c')", A) ;
		p[20] = x ;
		fputc('\n', stderr) ;
	    }
	    nref++;
	}
    }
    if (year3 && ((year3 < year[0]) || (year3 > year[1]))) nw++, fprintf(stderr,
	"++++ line %d: years not compatible (%d vs %d-%d)\n",
	lineno, year3, year[0], year[1]) ;
    if (err[phase]) nw++, fprintf(stderr, 
    	"++++ line %d: %s missing in '%s'++++\n", lineno, err[phase], fn);
    return(nw);
}

static int check_obj(ps)
/*++++++++++++++++
.PURPOSE  Verify the consistency of the "Objects" section
.RETURNS  Number of errors + warnings
-----------------*/
  SECTION *ps;	/* IN: The section to check */
{
  char *b, *s, *p, *eoh, last_char ;
  int nw = 0 ;
  int otext = 0 ;
  int nobj = 0 ;
  int lineno ;

    eoh = ps->ebody ; last_char = *eoh ; *eoh = 0 ;
    lineno = ps->lineno + 2 ;
    b = ps->abody ;

    /* Find the titles */
    while(*b) {
	if (*b == '\n') { lineno++; b++ ; continue ; }
	p = b; while (*p == ' ') p++ ;
	if (isdigit(*p)) break ;	/* Data line */
	s = b ; b += strloc(b, '\n'); 
	if (*p == '-') continue ;	/* Skip Rule */
	if (*p != 'R') {
	    nw++; fprintf(stderr, 
	    "++++ line %d: Title '   RA   (2000)  DE' missing?\n", lineno) ;
	    continue ;
	}
	while (isgraph(*p)) p++ ;	/* Skip "RA"	*/
	while (*p == ' ') p++ ;
	if (*p != '(' /*)*/) nw++, fprintf(stderr,
	    "++++ line %d: Equinox '(2000)' not in object list?\n", lineno) ;
	else while (isgraph(*p)) p++ ;	/* Skip Equinox	*/
	while (*p == ' ') p++ ;
	while (isgraph(*p)) p++ ;	/* Skip "Dec"	*/
	while (*p == ' ') p++ ;
	otext = p - s ;			/* Position of "Designations" */
	break ;
    }
    if (otext == 0) nw++, fprintf(stderr,
	"++++ line %d: no title in 'Objects' section?\n", ps->lineno) ;

    /* Verify the proper alignment of objects */
    while(*b) {
	if (*b == '\n') { lineno++; b++ ; continue ; }
	p = b; while (*p == ' ') p++ ;
	s = b ; b += strloc(b, '\n'); 
	if (*p == '-') continue ;	/* Skip Rule */
	nobj++ ;
	p = s + otext ;
	if ((p < b) && isgraph(*p) && p[-1] == ' ') continue ;
	nw++, fprintf(stderr, 
	    "++++ line %d: object misaligned (versus title)?\n", lineno) ;
    }

    if (nobj == 0) nw++, fprintf(stderr,
	"++++ line %d: 'Objects' section doesn't include any object??\n",
	ps->lineno) ;
    else fprintf(stderr, "---- %d objects declared in 'Objects' section\n", 
	nobj) ;
    *eoh = last_char ;
    return(nw) ;
}

/*==================================================================
		Signatures of Structures
 *==================================================================*/

void *cad_oid()
/*++++++++++++++++
.PURPOSE  Get a new OID
.RETURNS  the new oid
-----------------*/
{
   return(++free_oid);
}

char *cad_signature(name)
/*++++++++++++++++
.PURPOSE  Retrieve the exact signature
.RETURNS  The signature / NULL
.REMARKS  name by Control-C followed by 3-character 
-----------------*/
  char *name ;	/* IN: Structure name to find */
{
  int i ; char **a ;
    for (i=ITEMS(signatures), a=signatures; --i >= 0; a++) {
	if (strcmp(name, *a) == 0) return(*a) ;
    }
    return((char *)0) ;
}

int cad_tdup(cad)
/*++++++++++++++++
.PURPOSE  Duplicate the table descriptions if necessary
.RETURNS  0 (OK, was done) / 1 (done)
.REMARKS  The signature is changed (^C to ^D)
-----------------*/
  CATINFO *cad;		/* MOD: All loaded details	*/
{
  /* static char signature[] = "\04CAT"; */
  TABINFO **tlist, **alist, *tabf ;
  FILINFO *file; 
  int i, j ;

    if (cad->flags&CAT_tdup) return(0) ;

    alist = tlist = (TABINFO **)malloc(i = (cad->mtinfo+1)*sizeof(TABINFO *)) ;
    memset(alist, 0, i);
    for (file = cad->afinfo, j=cad->nfinfo ; --j >= 0; file++) {
	if (!file->tabform) continue ;
	for (tlist=alist; *tlist && (*tlist != file->tabform) ; tlist++) ;
	if (*tlist) {		/* Duplicate Table Description */
	    tabf = cad->atinfo + cad->ntinfo ;
	    cad->ntinfo += 1 ;
	    tbf_cpy(tabf, file->tabform) ;
	    file->tabform = tabf ;
	}
	else *tlist = file->tabform ;
    }
    free(alist) ;
    cad->flags |= CAT_tdup;
    return(1) ;
}

/*==================================================================
		Dealing with CATINFO Structures
 *==================================================================*/

CATINFO *cad_new(file, fop)
/*++++++++++++++++
.PURPOSE  Load the "format" File
.RETURNS  Loaded format / NULL if error
.REMARKS  The whole file is loaded in a single run.
-----------------*/
  char *file;	/* IN: Format File */
  int  fop;	/* IN: Options	   */
{
  struct stat stat_buf ;
  CATINFO *T; 
  FILINFO *pf, *pe;
  TABINFO *pt, *et, *qt; 
  COLINFO *pc, *qc ;
  SECTION *ps, *pSummary;
  char *p, *q, **a;
  char *tofree = (char *)0;
  int ftf ;			/* flag title found */
  char *e, *eol;
  int fh=0, i, j, f_op, lineno, errors, len;
  int4 count;
  short secno[100] ;
  /* int passes = 0; */

    /* Load file into Memory */
    time(&stat_buf.st_mtime) ;	/* Current date */
    if (!file) file = "" ;
    if ((file[0] == '-') && (file[1] == '\0')) file = "" ;
    if (*file) {	/* True File ... */
	i = stat(file, &stat_buf);
	if (i < 0) { perror(file); return((CATINFO *)0); }
	if (S_ISDIR(stat_buf.st_mode)) {
	    tofree = malloc(strlen(file) + 12);
	    strcpy(tofree, file);
	    /* V5.75: Try with "readme.viz" file if exists */
	    i = strlen(file);
	    strcpy(tofree+i, "/readme.viz");
	    if (stat(tofree, &stat_buf) < 0) 
		strcpy(tofree+i, "/ReadMe");
	    file = tofree;
	}
	if (fop&FOP_v) fprintf(stderr, "....Using ReadMe_file: %s\n", file);
        fh = open(file, O_BINARY);
        if (fh < 0) { 
	    perror(file); 
	    if (tofree) free(tofree); 
	    return((CATINFO *)0); 
	}
        count = lseek(fh, 0L, 2); lseek(fh, 0L, 0);
	fstat(fh, &stat_buf) ;
    }
    else count = load_stdin() ;
    T = (CATINFO *)malloc(count+sizeof(CATINFO)+strlen(file)+2);
    memset(T, 0, sizeof(CATINFO));
    memset(secno, 0, sizeof(secno)) ;
    T->signature = signatures[3] ;
    T->oid  = cad_oid();
    T->fbuf = (char *)(T+1);
    T->ebuf = T->fbuf + count ;
    T->credate = stat_buf.st_mtime ;
    if (*file) {
	i = read(fh, T->fbuf, count); 
        close(fh);
	/* Load =obsolete= */
	p = malloc(12+strlen(file)) ;
	strcpy(p, file);
	strcpy(p+1+strbloc(p,'/'), "=obsolete=");
	if ((T->more = load_file(p, 0))) T->flags |= CAT_obs ;
	free(p) ;
    }
    else load_move(T->fbuf) ;

#if 0
    /* Count the number of {}, realloc if any found */
    i=0;
    if (fop&FOP_cgi) for (p=T->fbuf; p<T->ebuf; p++) switch(*p) {
	case '{': case '}':	i++ ; continue ;
	case '\\': i += 2; continue ;
    }
    if (i > 0)	{	/* Add the escape characters */
	Tt = T;
	T = (CATINFO *)malloc(count+sizeof(CATINFO)+strlen(file)+2+i);
	memset(T, 0, sizeof(CATINFO));
    	T->signature = signatures[3] ;
	T->fbuf = (char *)(T+1);
	for (t=Tt->fbuf, p=T->fbuf; t<Tt->ebuf; p++,t++) switch(*t) {
	    case '\\': *(p++) = '\\'; *(p++) = 'b'; *p = ' '; continue ;
	    case '{': case '}': *(p++) = '\\' ; 	/* NO BREAK */
	    default:   *p = *t ;
	}
	T->ebuf = p;
	free(Tt) ;
    }
#endif

    /* Verify a pure ascii text ... */
    for (count=i=0, lineno=1, p=T->fbuf; p < T->ebuf; p++) {
	if (*p == '\n') lineno++;
	if (isascii(*p)) continue;
	count++; 
	if ((i<ITEMS(secno)) && (secno[i]!=lineno)) {
	    if (secno[i]) i++;
	    if (i<ITEMS(secno)) secno[i] = lineno;
	}
    }
    if (count) {	/* There are non-ascii chars */
	fprintf(stderr, "#+++ %d non-ascii chars (line", count);
	for (i=0; (i<ITEMS(secno)) && secno[i]; i++) {
	    if (i==0) p = secno[1] ? "s " : " ";
	    else p = ",";
	    fprintf(stderr, "%s%d", p, secno[i]);
	}
	fprintf(stderr, ")\n");
    }

    /* Remove trailing spaces & rules in text */
    T->ebuf = T->fbuf + str_text(T->fbuf, T->ebuf - T->fbuf);
    *(T->ebuf) = 0;
    T->ffile = T->filename = T->ebuf + 1; 
    strcpy(T->filename, file);
    T->ffile += strbloc(T->filename, dirsep) + 1;	/* Name without path */

	/* Count the number of sections in File */
	/* (a section header is not indented and contains a colon) */
    lineno = 0;
    memset(secno, 0, sizeof(secno)) ;
    for (count=0, p=T->fbuf, ftf=0; p < T->ebuf; p += i+1) {
    	i = strloc(p, '\n');
	if (!ftf) ftf = isspace(*p);
    	if (ftf && ((*p == '(') ||
    	   (isupper(*p) && (strloc(p, ':') < i)))) count++;
    }
    	/* Allocate the Sections */
    T->msec = count+3;
    T->section[0] = ps = (SECTION *)malloc(T->msec*sizeof(SECTION));
    bzero(ps, T->msec*sizeof(SECTION)) ;
    T->nsec = 1;
    	/* Fill the Sections	*/
    pSummary = (SECTION *)0;
    ps->signature = signatures[4] ;
    ps->oid    = cad_oid();
    ps->header = (char *)0; 
    ps->abody  = T->fbuf;
    ps->lineno = 0;
    ps->class  = SEC_Title;
    ps->secno  = ++secno[SEC_Title] ;
    ps->used   = 0;
    p = T->fbuf + strloc(T->fbuf,'\n') + 1;
    for (ftf=0, lineno=1; p < T->ebuf; p += i+1, lineno++) {
    	i = strloc(p, '\n');
	if (!ftf) ftf = isspace(*p);
    	if (ftf && ((*p == '(') ||
    	   (isupper(*p) && (strloc(p, ':') < i)))) {
    	    ps->ebody = ps->abody + str_text(ps->abody, p-ps->abody);
    	    q = ps->ebody;		/* End of Previous Section */
	    if (!(q[-1])) q--;
    	    *q = 0;			/* Terminate previous section */
    	    ps->next  = ps+1;
    	    ps++; T->nsec += 1;
            ps->signature = signatures[4] ;
            ps->oid    = cad_oid();
    	    ps->header = p;
    	    ps->abody   = p+i+1;
    	    ps->lineno = lineno;
    	    for (q = p+i; isspace(*q); q--) *q = 0;	/* Terminate title */
    	    ps->class  = sec_class(p);
    	    if ((ps->class == SEC_FileSummary) && (!pSummary))
    	    	pSummary = ps;
	    ps->secno = ++secno[(ps->class)&0xff] ;
    	    ps->used = 0;
    	}
    }
    ps->ebody = T->ebuf;    /* The body of last section is generally missing */
    if (ps->ebody <= ps->abody) ps->abody = ps->ebody;
    ps->next  = (SECTION *)0;
    T->section[1] = ps;		/* Last section */
#if DEBUG
    printf("----List of the %d Sections headers -------------", T->nsec);
    inbyte = getchar();
    for (ps=T->section[0]; ps; ps = ps->next) { int n;
    	n = ps->ebody - ps->abody;
    	printf("    Class=%02d#%03d(used=%02X), bytes=%d, Line%3d: %s\n", 
	    ps->class, ps->secno, ps->used, n, ps->lineno, 
	    ps->header ? ps->header : "(NULL)");
    }
#endif

	/* Count the number of items in "File Summary" */
    count = 0;
    if (pSummary) for (p = pSummary->abody; p < pSummary->ebody; p += i+1) {
    	i = strloc(p, '\n');
    	if (isspace(*p)) continue;	/* Not a file name */
    	if (match_rule(p)) continue;
      	count++;
    }
#if DEBUG
   	printf("....There are %d files in File Summary:", count);
   	/* inbyte = getchar(); */
   	if (pSummary) {
   	    printf("<%s>\n", pSummary->abody);
    	}
#endif

    	/* Allocate space for FILINFO descriptions */
    T->mfinfo = (count+3) + (count==0);
    T->afinfo = pf = (FILINFO *)malloc(T->mfinfo*sizeof(FILINFO));
    bzero(pf, T->mfinfo*sizeof(FILINFO)) ;
    pf->signature = signatures[2] ;
    pf->oid   = cad_oid();
    T->sfinfo = pSummary;
    T->nfinfo = T->ntinfo = T->mtinfo = 0;

    /* Fill what's found in File Summary */
    f_op = 0 ;		/* Flag for refs.dat file		*/
    eol= e = (char *)0; /* End-of-line indicator 		*/
    if (pSummary) for (p = pSummary->abody, lineno=pSummary->lineno+1;
      p < pSummary->ebody;  p = e+1, lineno++) {
	i = strloc(p, '\n');
	eol = e; e = p+i;	/* eol indicates the end of the explanations */
    	if (isspace(*p)) continue;	/* Not a file name */
    	if (match_rule(p)) continue;
	if (eol) while (isspace(*eol)) *(eol--) = 0;
        pf->signature = signatures[2] ;
        pf->oid   = cad_oid();
    	pf->dfile = p;  for (q = p; isgraph(*q); q++) ; 
    	if (*q) *(q++) = 0; while (isspace(*q)) q++; 
    	pf->reclen = atoi(q); while(isgraph(*q)) q++;
    	while (isspace(*q)) q++; 
    	pf->records = atol(q); while(isgraph(*q)) q++;
    	while (isspace(*q)) q++; 
    	pf->text = q;
    	pf->note = (SECTION *)0;
    	if (*pf->text == '*') {	/* Note related to File */
	    if ((ps = sec_findnote(pSummary, pf->dfile))) 
	    	ps->used |= SECNOTE_used, pf->note = ps; 
	}
    	pf->tabform = (TABINFO *)0;
    	pf->lineno = lineno;
    	for (q=p-1; isspace(*q); q--);
	if (pf->dfile[0] == 'r') {
	    if ((strcmp(pf->dfile, "refs.dat") == 0)
	     || (strcmp(pf->dfile, "refs") == 0)) f_op |= F_isFK ;
	}
    	T->nfinfo += 1;
    	pf++;
    }
    else lineno = 0;
    /* if (eol) while (isspace(*eol)) *(eol--) = 0; */

	/* Look for note /Equinox as parenthesized number */
    for (pf = T->afinfo, i = T->nfinfo; --i >= 0; pf++) { 
	/* Paragaphs not allowed in "File Summary */
	if ((q = strstr(pf->text, "\n--------------"))) *q = 0 ;
	/* Look for epoch/equinox: this word MUST BE HERE */
	pf->equinox = pf->epoch = (char *)0 ;
	for (q=pf->text; *q; q += strloc(q, ' '))  {
	    while (isspace(*q)) q++ ;	/* Start at a new word */
	    while (ispunct(*q)) q++ ;
	    if (strmatch("[BJ][12][0-9][0-9][0-9] coo*", q)) break ;
	    if (tolower(*q) != 'e') continue ;
	    if (stumatch("epoch*", q) || stumatch("equinox*", q)) break;
	}
	a = epochs(q, 0);
	if (a[0]) pf->equinox = strdup(a[0]);
	if (a[1]) pf->epoch   = strdup(a[1]);
    	q = pf->text + strlen(pf->text) - 1; 
    	if (*q == ')' && isdigit(q[-1])) {	/* Numbered (<1000!) note */
    	    for (--q; isdigit(*q); q--) ;
    	    if ((*q == '(') && (atoi(q+1)<1000)) {
	    	ps = sec_findnote(pSummary, q);
	    	if (ps) ps->used |= SECNOTE_used, pf->note = ps; 
	    }
    	}
    }
    
#if DEBUG
    printf("----List of the %d files from File Summary -------------\n", 
      T->nfinfo);
    for (pf=T->afinfo, i=T->nfinfo; --i >= 0; pf++)
      printf(
 	"    Line#%d: File=%s reclen=%d records=%d tabform=%p text=\"%s\"\n",
    	pf->lineno+1, pf->dfile ? pf->dfile : "(?)", pf->reclen, pf->records,
    	pf->tabform,
    	pf->text ? pf->text : "(NULL)");
    printf("...Hit Return "); inbyte = getchar();
#endif

    /* Count number of "Byte-by-byte" */
    for (ps=T->section[0]; ps; ps = ps->next) {
    	if (ps->class == SEC_BytebyByte) count++;
    	if (ps->class == SEC_Col) count++;
    }
    if (count == 0) {	/* Assume first section is Byte-by-Byte */
    	ps = T->section[0];
    	ps->class = SEC_BytebyByte;
    	count = 1;
    }
    T->mtinfo = count + 1;
    if (T->mtinfo < T->mfinfo) T->mtinfo = T->mfinfo ;
    T->atinfo = pt = (TABINFO *)malloc(T->mtinfo*sizeof(TABINFO));
    bzero(pt, T->mtinfo*sizeof(TABINFO)) ;
    pt->signature = signatures[1] ;
    pt->oid = cad_oid();

    /* Fill the "Byte-by-byte" part */
    for (ps=T->section[0]; ps; ps = ps->next) {
    	if ((ps->class == SEC_BytebyByte) || (ps->class == SEC_Col)) ;
	else continue;
	for (pf = T->afinfo, pe = pf+T->nfinfo; pf < pe; pf++) {
	    if (pf->tabform) continue;
	    /* Connect to the FILINFO */
	    if (match_word(ps->header, pf->dfile, ':')) {
	    	ps->used |= SECNOTE_used; /* This Byte-by-Byte is used */
	    	pf->tabform = pt;
	    }
	}
	if (ps->class == SEC_Col) {	/* CSV not declared: default TSV */
	    if ((p = strstr(ps->header, "-separated"))) {
		--p;
		if (ispunct(*p)) pt->charsep = *p;
		else if (stumatch(ps->header, "*comma-sepa*"))
		  pt->charsep = ',';
		else if (stumatch(ps->header, "*TAB-separa*")) 
		  pt->charsep = '\t';
		else if (fop&FOP_quiet) ;
		else fprintf(stderr, 
		  "++++cad_new: can't interpret column separator in:\n  %s\n",
		  ps->header);
	    }
	    else if (strmatch(ps->header, "*.tsv"))
		  pt->charsep = '\t';
	    else pt->charsep = ',';
	}
	if (ps->ebody > ps->abody) {	/* Format is here */
    	    T->ntinfo += 1; 
            pt->signature = signatures[1] ;
            pt->oid = cad_oid();
    	    pt->section = ps;
    	    pt->lineno  = ps->lineno + (ps->header ? 1 : 0);
    	    pt->posfmt[0] = -1;		/* No position */
    	    pt->nfmt = 0;
	    pt->equinox = pt->epoch = (char *)0;
    	    pt->afmt = (COLINFO *)0;
	    /* V5.8:  accept comments, e.g.
		Byte-by-Byte Description (# are comments) of: 
		Byte-by-Byte Description (10 heading lines) of:
	    */
	    if ((p = ps->header)) p = strchr(p, '(');
	    if (p) {
		/* Indication of heading lines may be:
		   (# are comments) or (12 headlines)
		*/
		++p;
		i = strloc(p, ')');
		for (j=0; j<i; j++) { if (isdigit(p[j])) break; }
		if (isdigit(p[j])) pt->headlines = atoi(p+j);
		else if (ispunct(*p)) pt->headchar = *p;
		else fprintf(stderr, 
		  "++++cad_new: can't interpret headlines at line %d:\n  %s\n",
		  ps->lineno, ps->header);
	    }
    	    pt++;
    	}
	else if (ps->class == SEC_Col)	/* CSV not declared: default TSV */
	   pt->charsep = '\t';
    }
    if (T->ntinfo == 0) {	/* Just a single description */
    	T->ntinfo = 1; 
    	pt->section = T->section[0];
    	pt->nfmt = 0;
	pt->equinox = pt->epoch = (char *)0;
    	(T->afinfo)->tabform = pt;
    }
#if DEBUG
    printf("List of the %d Byte-by-Byte Descriptions-------------", T->ntinfo);
    printf("....Hit Return "); inbyte = getchar();
    for (pt=T->atinfo, i=T->ntinfo; --i >= 0; pt++) { SECTION *ps;
    	ps = pt->section;
    	printf("    Line#%d: %d columns, section#%d.%d '%s'\n", pt->lineno+1, 
    	    pt->nfmt, ps->class, ps->secno, ps->header? ps->header : "(NULL)");
    }
    printf("List of the %d Files-------------\n", T->nfinfo);
    for (pf=T->afinfo, i=T->nfinfo; --i >= 0; pf++) { 
    	printf("    Line#%d: File=%s reclen=%d records=%d format@%p\n",
    	pf->lineno+1, pf->dfile ? pf->dfile : "(?)", pf->reclen, pf->records,
    	pf->tabform );
    }
    printf("....Hit Return "); inbyte = getchar();
#endif

    /* For each described file, load format */
    for (i=T->ntinfo, pt=T->atinfo, errors=0; --i >= 0; pt++) 
    	errors += tbf_load(pt, fop, f_op);

    /* Install the Equinox+Epoch for each file, and verify reclen (V5.81) */
    for (i=T->nfinfo, pf=T->afinfo; --i >= 0; pf++) {
	pt = pf->tabform;
	if (!pt) continue;
	if (pt->nfmt == 0) {
	    if (strncasecmp(pf->dfile, "readme", 6)) fprintf(stderr, 
		    "++++table without columns: %s\n", pf->dfile);
	    continue;
	}
	if ((!pf->equinox) && (pt->equinox)) pf->equinox = strdup(pt->equinox);
	if ((!pf->epoch  ) && (pt->epoch  )) pf->epoch   = strdup(pt->epoch  );
	/* V5.81: adjust the reclen. */
	pc = &pt->afmt[pt->nfmt-1];	/* Last column */
	len = pc->b0 + pc->rep*pc->len;	/* V5.84 */
	if (len < pf->reclen) {		/* There are extra blanks ... */
	    if (fop&FOP_quiet) ;
	    else fprintf(stderr, 
	      "#...Lrecl of file '%s' (%d) indicates %d extra blanks (trim!)\n",
	      pf->dfile, pf->reclen, pf->reclen - len);
	}
	else if (len > pf->reclen){	/* No record has the max. length */
	    if (fop&FOP_quiet) ;
	    /* if reclen==0 ==> it's generally ok */
	    else if (pf->reclen) fprintf(stderr, 
	      "++++Lrecl of file '%s' (%d) increased to %d\n", 
	      pf->dfile, pf->reclen, len);
	    pf->reclen = len; 
	}
    }

    /* Install the FK / PK for the first column of tables (from 2nd) */
    pe = T->afinfo + T->nfinfo;
    for (pt=T->atinfo, et=pt+T->ntinfo; pt < et ; pt++) {
	pc = pt->afmt ;		/* 1st column */
	if (!pt->nfmt) continue; /* No column */
	if (!pc) continue ;
	ftf = 0 ;		/* Set when a column is found */
	/* Find a file having this description */
	for (pf = T->afinfo; (pf<pe) && (pf->tabform != pt); pf++) ;
	if (pf == pe) continue;
	for (qt=T->atinfo; qt<et; qt++) {
	    /* passes++; */
	    if (pt == qt) continue ;
	    qc = tbf_findlabel(qt, pc->label) ;	/* Column of same name */
	    if (!qc) continue ;	
	    if (pc->ff != qc->ff) continue ;
	    if (pc->len != qc->len) continue ;
	    if (pc->odp != qc->odp) continue ;
	    if (strcmp(pc->unit, qc->unit)) continue ;
	    if (strword(qc->text, pf->dfile)) {
		/* Column says e.g. 'table' */
		ftf += 1 ,
		qc->flags |= F_isFK ;
	    }
	}
	if (ftf) 		/* Column is used, is threfore PK */
	    pt->afmt->flags |= F_isPK ;
    }
    /* fprintf(stderr, "....passes for testing PF/FK=%d\n", passes); */

    /* Individualize the TABINFO for each file  */
    if (fop&FOP_tdup) cad_tdup(T) ;

#if DEBUG
    printf("List of the %d Sections headers -------------", T->nsec);
    inbyte = getchar();
    for (ps=T->section[0]; ps; ps = ps->next) { int n;
    	n = ps->ebody - ps->abody;
    	printf("    Class=%02d#%03d(used=%02X), bytes=%d, Line%3d: %s\n", 
	    ps->class, ps->secno, ps->used, n, ps->lineno, 
	    ps->header ? ps->header : "(NULL)");
    }
    printf("List of the File Summary: -------------", T->nsec);
    inbyte = getchar();
    for (pf = T->afinfo, i=T->nfinfo; --i >= 0; pf++) {
    	if (! pf->dfile) continue;
    	printf("%12s %4d %8d \"%s\"\n", pf->dfile, pf->reclen, pf->records, 
    	    pf->text);
      	if (pf->note) printf("----%s\n%s\n", pf->note->header, pf->note->abody);
      	if (pt = pf->tabform) printf("          @%p: %s\n", 
      	    pt, pt->section->header);
    }
#endif
    if (errors) {
	if (fop&FOP_v) {	/* Verbose option, edit the Sections */
	    printf("====Section Summary of File %s:\n", p);
    	    for (ps=T->section[0]; ps; ps = ps->next) sec_prhead(ps, stdout);
	}
    	cad_kill(T);
    	T = (CATINFO *)0;
    }

    if (tofree) free(tofree);
    return(T);
} 

int cad_kill(T)
/*++++++++++++++++
.PURPOSE  Free an allocated Format structure
.RETURNS  0 (OK) / -1 (error)
-----------------*/
  CATINFO *T;	/* IN: Format to free */
{
  TABINFO *pt; FILINFO *pf; 
  int it;

    if (!T)	return(-1);	/* Already free */
#if DEBUG
    osmtrace(1);
#endif

    if (!T->signature) {
        fprintf(stderr, "****cad_kill: can't free unsigned object 0x%p\n", T) ;
	return(-1) ;
    }
    if (T->signature != signatures[3])  	/* Other structure... */
        return(tbf_free((TABINFO *)T)) ;
    
    /* Free first the COLINFO */
    for (pt=T->atinfo, it=T->ntinfo; --it>=0; pt++)
    	tbf_free(pt);
    /* Free equinoxes */
    for (pf=T->afinfo, it=T->nfinfo; --it>=0; pf++) {
	if(pf->equinox) free(pf->equinox);
	if(pf->epoch)   free(pf->epoch);
	if(pf->comment) free(pf->comment);
	if(pf->more) 	free(pf->more);
	if(pf->flags&F_Atext) free(pf->text) ;
	/* F_Aunit flag in FILINFO indicates the TABform was allocated */
	if(pf->flags&F_Aunit) free(pf->tabform) ;
	pf->signature = (char *)0;
    }
    
    /* Free the other arrays */
    if (T->atinfo)     free(T->atinfo);
    if (T->afinfo)     free(T->afinfo);
    if (T->section[0]) free(T->section[0]);
    if (T->more)       free(T->more);
    if (T->comment)    free(T->comment);
    /* printf("....will free CATINFO\n");	*/
    T->signature = (char *)0 ;
    free(T);
    catid[0] = 0;	/* catid not known any more	*/
    return(0);
} 

int cad_check(T)
/*++++++++++++++++
.PURPOSE  Check various consistencies
.RETURNS  Number of warnings
-----------------*/
  CATINFO *T;	/* IN: Loaded ReadMe to check */
{
  SECTION *ps; FILINFO *pf, *pIntro; TABINFO *pt, *et; 
  char buft[256] ;	/* To check unused byte-by-Byte */
  int i, t, nw, nit=0;

    /* Verify the existence of TABs in ReadMe file */
    i = strloc(T->fbuf, '\t');
    if (T->fbuf[i]) fprintf(stderr, 
	    "++++Warning: there are tab(s) in file '%s'\n", T->ffile);

    /* Verify the "Objects" section */
    for (ps = T->section[0]; ps ; ps = ps->next) {
	if (ps->class == SEC_Objects) check_obj(ps) ;
	if (ps->class == SEC_SeeAlso) check_see(ps) ;
    }

    /* Verify the Title */
    nw = check_title(T->section[0], T->ffile);

    /* Verify the ADC_Keywords */
    for (ps = T->section[0]; ps ; ps = ps->next) {
	if (ps->class != SEC_Keywords) continue ;
	if (starts_by(ps->header, "ADC")) {
	    nw += check_keywords(ps, T->ffile);
	    break ;
	}
    }
    if (!ps) fprintf(stderr, "++++ Missing ADC_Keywords in file '%s'\n", 
	T->ffile);

    /* Verify the list of Files: the Intro file exists, notes are there:
     * buft is used here to mark the descriptions used.
     * nit counts the number of ignored tables. */
    memset(buft, 0, sizeof(buft)) ;
    pIntro = (FILINFO *)0;
    pt = T->atinfo ; et = pt + T->ntinfo ;
    for (pf = T->afinfo, i = T->nfinfo; --i>=0; pf++) {
    	if (strdiff(T->ffile, pf->dfile) == 0) pIntro = pf;
	if (pf->flags&F_NODB) nit++;
    	nw += check_file(pf, T->ffile);
    	if ((*pf->text == '*') && (!pf->note)) nw++, 
    	    fprintf(stderr, 
	    "++++ line %d: note (*) for file '%s' not found ++++\n",
	    pf->lineno+1, pf->dfile);
	while ((pt<et) && (pt != pf->tabform)) pt++ ;
	if (pt>=et) for (pt = T->atinfo; (pt<et) && (pt != pf->tabform); pt++) ;
	if (pt<et) {
	    t = pt - T->atinfo ;
	    if (t < sizeof(buft)) buft[t] = 1 ;	/* Mark as Used */
	}
    }
    if (!pIntro) nw++, fprintf(stderr, 
    	"++++ File '%s' missing in File Summary\n", T->ffile);
    	
    /* Verify the correctness of each "Byte-by-Byte Description":
     * the corresponding buft is still zero */
    for (pt=T->atinfo, et=pt+T->ntinfo, t=0; pt<et ; pt++, t++)  {
    	nw += tbf_check(pt);
	if ((buft[t] == 0)&&((pt->flags&F_NODB)==0)) fprintf(stderr,
	    "++++ line %d: unused <%s>\n", 
	    pt->lineno+1, pt->section->header ? pt->section->header :
	      "Byte-by-Byte description");
	    /*"++++ line %d: Byte-by-Byte description not used\n", 
	    pt->lineno+1); */
#if 0
	/* Verify the Primary Key / Foreign Key */
	for (pc=pt->afmt, ec=pc+pt->nfmt; pc<ec; pc++) {
	    if (!(pc->flags & F_isFK)) continue ;
	}
#endif
    }
    /* Find out the unused Byte-by-Byte description */

    
    /* Find out the unused notes */
    for (ps=T->section[0]; ps; ps=ps->next) {
#if DEBUG
	int n;
    	n = ps->ebody - ps->abody;
    	printf("    Class=%02d#%03d(used=%02X), bytes=%d, Line%3d: %s\n", 
	    ps->class, ps->secno, ps->used, n, ps->lineno, 
	    ps->header ? ps->header : "(NULL)");
#endif
    	if (ps->used & SECNOTE_used) continue;
    	if ((ps->class == SEC_Note)|| (ps->class == SEC_BytebyByte)) nw++,
	    fprintf(stderr, "++++ line %d: '%s' never used ++++\n", 
	    ps->lineno+1, ps->header ? ps->header : "(NULL)");
    }

    /* Verify finally the "End" Section */
    ps = T->section[1];
    if (ps->class != SEC_End) nw++, fprintf(stderr,
    	"++++(End) last line is missing in '%s'++++\n", T->ffile);

    return(nw);
}

static int pr_Section(ps, fout)
/*++++++++++++++++
.PURPOSE  Print out a Section
.RETURNS  2
.REMARKS  catid[0] not zero indicates a CGI edition
-----------------*/
  SECTION *ps;		/* IN: Section to Edit		*/
  FILE   *fout;		/* IN: Opened output file 	*/
{
  char *p, *e, x;
    /* printf("....pr_Section, class=%d, Tbf=%s\n", ps->class, Tbf); */
    if ((p = ps->header)) {
    	for (e=p; *e && (*e != ':'); e++) ;
    	if (Tbf && *Tbf) {
	    switch(ps->class) {
	      case SEC_End:
	      case SEC_Title:
		for (e=p; isgraph(*e); e++) ;
		break;
	      case SEC_FileSummary:
	      case SEC_Keywords:
		break ;
	      default:
    	    	fprintf(fout, "\\label{%s}", sec_id(ps)) ;
		break;
	    }
    	    fprintf(fout, "%s", Tbf);
    	}
    	while (p < e) fputc(*p, fout), p++;
    	if (Trm) fprintf(fout, "%s", Trm);
    	while (*p) fputc(*p, fout), p++;
    	fputc('\n', fout);
    }

    p = ps->abody;
    /* Note that Title Section has no header...	Write CATID in \bf   */
    if ((ps->class == SEC_Title) && (!ps->header) && Tbf && *Tbf) {
	fprintf(fout, "%s\\plain{", Tbf);
	while (isgraph(*p)) (*FPUTC)(*p, fout), p++;
	fprintf(fout, "}%s", Trm);
    }
    /* We have, in the "See Also" section, to enclose the catalogue
       designations between < > (only for HTML ouput)
    */
    if ((ps->class == SEC_SeeAlso) && Tbf) {
	while (*p){
	    while (isspace(*p)) (*FPUTC)(*p, fout), p++;
	    if (isanchor(p)) {
	        fputc('<', fout);
	        while (isgraph(*p)) (*FPUTC)(*p, fout), p++;
	        fputc('>', fout);
	    }
	    while (*p && (*p != '\n')) (*FPUTC)(*p, fout), p++;
	}
        fputc('\n', fout);
    }
#if 0
    if (p < ps->ebody) fprintf(fout, "%s\n", p);
#else
    if (Tbf) {	/* Print with link generation */
	x = *(ps->ebody) ; *(ps->ebody) = 0 ;
	print_cgi(p, fout) ;
	*(ps->ebody) = x ;
        fputc('\n', fout) ;
    }
    else {
	while (p < ps->ebody) (*FPUTC)(*p, fout), p++;
        fputc('\n', fout) ;
    }
#endif
    return(2);
}

static int pr_Summary(pf, fout, P_opt)
/*++++++++++++++++
.PURPOSE  Print out the "File Summary"
.RETURNS  1
.REMARKS  The Tget / Tbf / Trm indicate the CGI edition
-----------------*/
  FILINFO *pf;		/* IN: The parameters to print 	*/
  FILE   *fout;		/* IN: Opened output file 	*/
  int P_opt;		/* IN: Printing Option		*/
{
  char *p; int i;
  SECTION *ps ;

    if (*(pf->dfile) == 0177) return(0);
    if (Tget[0]) {
	fprintf(fout, "\\label{%s}", sec_id(pf));
#if FETCH
	if (P_opt > 1) fprintf(fout, "\\boxa{%s} ", pf->dfile);
#else
	if (P_opt > 1) fprintf(fout, "{\\times} ") ;
#endif
    }
    if (Tget[1] && pf->tabform) {
	ps = pf->tabform->section ;
	fprintf(fout, "\\yAref{Byte-by-byte Description of %s}{%s}{", 
		pf->dfile, sec_id(ps)) ;
	/*fprintf(fout, "\\Aref{%s}{", sec_id(ps)) ;*/
    }
    fprintf(fout, "%s", pf->dfile); i = 14 - strlen(pf->dfile);
    if (Tget[1] && pf->tabform) fputc('}', fout);
    while (--i >= 0) fputc(' ', fout);
    fprintf(fout, " %5d", pf->reclen);
    if (pf->records == 0) fprintf(fout, " %8s", ".");
    else fprintf(fout, " %8d", pf->records);
    p = pf->text ? pf->text : "" ;
    fprintf(fout, "  "); if (*p != '*') fputc(' ', fout);
    if (Tget[0] && pf->note) {
    	i = *p;
    	if (i) p++;
	ps = pf->note ;
    	fprintf(fout, "\\yAref{%s}{%s}{%c}%s\n", 
		"move to details of note", sec_id(ps), i, p);
    	/*fprintf(fout, "\\Aref{%s}{%c}%s\n", sec_id(ps), i, p);*/
    }
    else fprintf(fout, "%s\n", p);
    return(1);
}

int cad_print(T, fout, P_opt)
/*++++++++++++++++
.PURPOSE  Print out the "ReadMe" file
.RETURNS  Number of lines
-----------------*/
  CATINFO *T;		/* IN: Various details about position of description */
  FILE   *fout;		/* IN: Opened output file 	*/
  int P_opt;		/* IN: Printing Option		*/
{
static char intro[] = "\
\\cgidef{-pre -tex2 -bib -cat -mail -oglu\n\
\\Unit\t#1{\\W{explain}{%s/cgi-bin/Unit?\\htarg{#1}}{#1}}\n\
\\boxa\t#1{\\tag{INPUT TYPE=\"checkbox\" NAME=\"%s\" VALUE=\"#1\"}}\n\
}\n" ;
static char choice[] = "\\boxa{=ALL=} All tables   \
\\begin{SELECT NAME=\"fmt\"}\
\\option{ascii}\
\\option{TAR}\
\\option{FITS}\
\\end{SELECT}\
\\tag{INPUT TYPE=\"submit\" VALUE=\" Fetch File(s) \"}\
\\end{FORM}\n\\hr\n"; 

  SECTION *ps; FILINFO *pf, *pReadMe; TABINFO *pt; char *getenv();
  char *p, last_class=0; int i, lines, notes;
  static FILINFO ReadMe = { "\3FIL", (char *)0, (void *)0, 80, 0 } ;
  char filename[64], *pn;
  char *aroot ;
  
    catinfo = T ;
    aroot = P_opt&P_a ? aanchor : "" ;
    FPUTC = fputc ;
    lines = 0;
    if (!ReadMe.text) ReadMe.text = "This file" ;
    ReadMe.dfile = T->ffile;
    if (P_opt) {			/* Print cgidef	*/
    	set_catid(T->section[0], T->ffile);
    	fprintf(fout, intro, aroot, catid) ;	/* Definitions	*/
    	Tbf = "{\\bf ", Trm = "}"; /* FPUTC = tex_fputc ; */
    	Tget[0] = Tget[1] = ' ';
    	rule_char = '\\';	/* TeX Rule		*/
    }
    else {
    	catid[0] = 0; 
    	Tbf = Trm = "";
    	Tget[0] = Tget[1] = 0;
    	rule_char = '-';
    }

    /* Print out Titles, until the File Summary		*/
    for (ps=T->section[0]; ps; ps=ps->next) {
    	if (ps->class > SEC_End) fputc('\n', fout);
	/* if (ps->class != SEC_Title) fputc('\n', fout); */
    	if (ps->class == SEC_FileSummary) break;
    	if (ps->class == SEC_BytebyByte) break;
    	lines += pr_Section(ps, fout);
	if (ps->class == SEC_Title) fprintf(fout, "%s\n", rule('='));
    }
#if DEBUG
	inbyte = getchar();
#endif

    	/* Edit the "File Summary" */
    fprintf(fout, "\n%sFile Summary%s:\n%s\n", Tbf, Trm, rule(rule_char)); 
    if (Tget[0]) {
	fprintf(fout, "      ") ;
#if FETCH
	if (P_opt > 1)
	fprintf(fout, "\\begin{FORM METHOD=POST ACTION=\"GetFile\"}\n");
#endif
    }
    fprintf(fout, "%-13s %6s %8s   %s\n%s\n", 
    	" FileName", "Lrecl", "Records", "Explanations", rule(rule_char));
    pReadMe = cad_findfile(T, T->ffile);
    if (!pReadMe) pReadMe = &ReadMe;
    pr_Summary(pReadMe, fout, P_opt); pReadMe->dfile = "\177";
    notes = 0;			/* notes tells associated notes */

    for (i=T->nfinfo, pf=T->afinfo; --i >= 0; pf++) 
    	lines += pr_Summary(pf, fout, P_opt);
    pReadMe->dfile = T->ffile;
    fprintf(fout, "%s\n", rule(rule_char)); fflush(fout);
#if FETCH	/*==== For Later ====*/
    if (Tget[0] && (P_opt > 1)) {
	fprintf(fout, "%s", choice) ;
    }
#endif
#if DEBUG
	inbyte = getchar();
#endif
    
    for(pt = T->atinfo; ps; ps = ps->next) {	/* Write all other sections */
	/* sec_prhead(ps, fout); */
    	if (ps->class == SEC_FileSummary) continue;
    	if (ps->class == SEC_Note) notes += 1;
	else if (notes) {	/* Terminates notes after tables */
    	    fprintf(fout, "%s\n", rule(rule_char));
	    notes = 0;
	}
    	last_class = ps->class;
    	if (last_class == SEC_BytebyByte) {
    	    p = ps->header;
    	    if (p) p += strloc(p, ':'); else p = ":";
    	    if (*p) p++;
    	    while (isspace(*p)) p++;
    	    fputc('\n', fout);
    	    if (*Tbf) fprintf(fout, "\\label{%s}", sec_id(ps)) ;
    	    fprintf(fout, "%sByte-by-byte Description of file%s", Tbf, Trm);
	    if (pt->headchar) fprintf(fout, " (%c headlines)", pt->headchar);
	    else if (pt->headlines) fprintf(fout, " (%d headlines)", 
		    pt->headlines);
	    fputc(':', fout);
    	    if (*p) while (*p) {	/* Write the names of the files	*/
		for (pn=filename; isgraph(*p); p++, pn++) *pn = *p;
		*pn = 0;
		/*if (*Tbf) fprintf(fout, " \\Aref{%s}{", 
		    sec_id(cad_findfile(T, filename))); */
		if (*Tbf) fprintf(fout, " \\yAref{Summary of %s}{%s}{", 
		    filename, sec_id(cad_findfile(T, filename)));
		fprintf(fout, "%s", filename); 
		if (*Tbf) fputc('}', fout);
		while (isspace(*p)) p++;
	    }
    	    else for (i=T->nfinfo, pf=T->afinfo; --i >= 0; pf++) {
    	    	if ((pf->tabform)->section == ps) 
    	    	   fprintf(fout, " %s", pf->dfile);
    	    }
	    fflush(fout);
#if DEBUG
	    inbyte = getchar();
#endif
	    /* NOTE: There may be several consecutive 
	       "Byte-by-byte Description" sections, the (n-1) first
	       having no descrition (ps->ebody == ps->abody)
	       When this happens, there is no blank line between
	       the consecutive "Byte-by-byte Description" sections.
	    */

	    if (ps->ebody > ps->abody) { 	/* There is a format */
    	    	fprintf(fout, "\n");
    	    	lines += tbf_print(pt, fout, P_opt); fflush(fout);
    	    	pt++;
	    }
    	    continue;
    	}
    	if (ps->class > SEC_End) fputc('\n', fout);
    	if (ps->class == SEC_End) fprintf(fout, "%s\n", rule('='));
    	lines += pr_Section(ps, fout);
	fflush(fout);
    }

    	/* Write the "End" */
    if (last_class != SEC_End) { 
    	p = getenv("VIme");
    	if (!p) p = getenv("USER");
    	if (!p) p = "[CDS?]";
    	fprintf(fout, "%s\n(End)", rule('='));
    	for (i=(75-11-2)-strlen(p); --i >= 0; ) fputc(' ', fout);
    	fprintf(fout, "%s  %s\n", p, date(0));	/* User Name */
    }
    
    return(lines);
}

#if TEST
/*==================================================================
		Test Program
 *==================================================================*/
main (argc, argv)
/*++++++++++++++++
.PURPOSE  Just a TEST program
-----------------*/
  int argc; 
  char **argv;
{
  char *p; CATINFO *pt; int fop, i; FILINFO *pf ;

    if (argc < 2) {
	printf("Usage: catstd ReadMeFile [options]\n");
	exit(1) ;
    }
    p = argc > 2 ? argv[2] : "" ;
    for (fop=0; *p; p++) switch(*p) {
    	case '.': fop |= FOP_dot; continue;
    	case '1': fop |= FOP_1  ; continue;
    	case 'w': fop |= FOP_w  ; continue;
    	case 's': fop |= FOP_std; continue;
    	case 'v': fop |= FOP_v  ; continue;
    	case 'r': fop |= FOP_ra ; continue;
    	defaut: fprintf(stderr, "++++Option '%c' ignored++++\n", *p);
    }
    pt = cad_new(argv[1], fop);
    if (!pt) exit(1);
    printf("\n====Will check===="); inbyte = getchar();
    cad_check(pt);
    for (i=pt->nfinfo, pf=pt->afinfo; --i >= 0; pf++) {
	printf("File %12s:", pf->dfile);
	if (pf->equinox) printf(" equinox=%s", pf->equinox);
	if (pf->epoch  ) printf("   epoch=%s", pf->epoch  );
	printf("\n");
    }
    printf("\n====Will edit the file===="); inbyte = getchar();
    cad_print(pt, stdout, 0);
    printf("\n====Will edit the file for cgi===="); inbyte = getchar();
    cad_print(pt, stdout, P_cgi);
    printf("\n====Will Kill everything===="); inbyte = getchar();
    cad_kill(pt);
}
#endif
