/*++++++++++++++
.IDENTIFICATION fits.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS
.VERSION  1.0   15-Jul-1993: C-shell version(mods 1.1 to 1.9 deleted)
.VERSION  2.0   02-Sep-1994: C package 
.VERSION  2.2   08-Dec-1994: Take care of long strings
.VERSION  2.3   04-Feb-1995: fit_prkn: keep only ONE word
.VERSION  2.4   07-Mar-2001: Usage in XML2FITS Context.
.VERSION  2.5   14-Mar-2001: fits_flushnl
.VERSION  2.6   02-Jun-2007: Cosmetics (-Wall)
.VERSION  2.7   15-Dec-2009: 64-bit
.COMMENTS       Utilities for Asci Fits i/o
	Note that the toFITSd local variable, which can be modified
	by fit_setmode, specifies the insertion of newlines
	(when toFITSd is zero)
	Note also that an array 'fpos' of file positions (modulo 2880!) is
	kept here, because the "ftell" function doesn't seem to work
	on pipes or non-disk files... but "ftell" does work on PC!!
---------------*/

#define VERBOP 0

#ifndef DEBUG
#define DEBUG 0
#endif

#ifdef XML2FITS
#include <stdio.h>
#include <string.h>
#include <ctype.h>
int match_rule(), match_num(), strnline(), strloc();
#else
#include <catstd.h>
#include <filutil.h>
#endif
#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])

#if DEBUG
static char inbyte;
#endif

static char toFITSd = 1;	/* 0 = don't block, 1 = block */
static long fpos[20];		/* File position for streams  */

    	/* The following static variables are used to avoid
    	   parameter passing to internal subroutines
   	*/
static FILE *f;			/* The file being used		*/
static char noslash = 0;	/* indicator in comment()	*/
static char fill_char = ' ';	/* Char used in fill()		*/
static int4 NAXIS[3];		/* Keep NAXIS read in fit_prfile*/

/* Accept 8-bit charactesr */
#define isGraph(c)	(c && (!isspace(c)))

/*==================================================================
		Internal utilities
 *==================================================================*/
static int fill(o)
/*++++++++++++++++
.PURPOSE  Completes the curent line
.RETURNS  1
-----------------*/
  int o;	/* IN: Current position in line */
{
  int  i;
#if DEBUG>1
	printf("....fill(%d)\n", o);
#endif
    if (toFITSd || fill_char != ' ') 
    	for (i=o; i<80; i++) fputc(fill_char, f);
    else fputc('\n', f);
    if (fileno(f) < ITEMS(fpos)) fpos[fileno(f)] += 80;
    return(1);
}

static int prkw(kw, eq)
/*++++++++++++++++
.PURPOSE  Issue COMMENT
.RETURNS  Number of bytes written
-----------------*/
  char *kw;	/* IN: Keyword to write out	*/
  int  eq;	/* IN: Option 1 to append =	*/
{
  int i; char *p;

    if (!kw) kw = "";
    /* printf("....calling prkw(\"%s\", %d), fout=%d\n", kw, eq, fileno(f));
	-- Write out the keyword name (limited to 8 bytes) */
    for (i=0, p=kw; (*p) && (i<80); i++, p++) fputc(*p, f);
    while (i < 8) fputc(' ', f), i++;
    if (eq) fputc('=', f), i++;
    fputc(' ', f), i++;
    return(i);
}

static int comment(o, com)
/*++++++++++++++++
.PURPOSE  Append comment to a keyword
.RETURNS  Number of lines
.REMARKS  The comment is continued on a second (blank) line if necessary
	Note that control characters (tabs, newlines, etc) are replaced by
	a blank.
-----------------*/
  int o;	/* IN: Current position in line */
  char *com;	/* IN: Comment to append	*/
{
  char *p, *w, x, cfill; int i, indent, lines, len;
#if DEBUG>1
    printf("....comment(%d, \"%s\"(%d bytes))\n", o, com, strlen(com));
#endif
    i = o;
    if ((noslash==0) && *com && (i > 18)) {	/* Add the comment symbol (/) */
    	if (i < 80) fputc(' ', f), i++;
    	while (i < 31) fputc(' ', f), i++;	/* '/' in col. 32 */
    	if (i < 80) fputc('/', f), i++;
    	if (i < 80) fputc(' ', f), i++;
    }
    indent = MIN(i, 33);			/* For continuations 	*/
    for (p=com, lines=0; *p; ) {	/* Write out comment lines 	*/
    	while (i < indent) fputc(' ', f), i++;	/* Add indentation	*/
    	x = 0;					/* Indicator for space	*/
    	while ((*p) && (i<80)) {		/* Write word-by-word 	*/
#if DEBUG>1
    	    printf("....Still %d bytes, i=%d", strlen(p), i);
#endif
    	    while (isspace(*p)) p++;
    	    for (w=p; isGraph(*w); w++) ; 
	    /* Accept 8-bit chars (are non-graph) */
	    while (*w && (!(isspace(*w)))) w++;
    	    len = w-p;				/* Length of next word	*/
    	    if ((len+x+i) > 80)	 {		/* Can't match on line	*/
    	    	if (x) break;			/* Will be on next line	*/
    	    	len = 80-i;			/* Word too long...	*/
	    }
	    if (x) fputc(' ', f), i++;		/* Blank separator	*/
    	    if (match_rule(p)) {			/* Add a rule	*/
    	    	cfill = fill_char; fill_char = *p;
    	    	lines += fill(i);  		/* Fill up current line	*/
    	    	fill_char = cfill;
    	    	i = 0; p = w;
    	    }
    	    else { 				/* Write current word	*/
    	    	fwrite(p, 1, len, f);		
    	    	p += len; i += len;
	    }
#if DEBUG>1
    	    printf("=>%d, bytes=%d\n", i, strlen(p));
#endif
	    x = 1;		/* Will insert space before next word	*/
    	    while (isspace(*p)) p++;
    	}
    	lines += fill(i);  			/* Fill up current line	*/
    	i=0;					/* Current line empty	*/
    }
    if (i) lines += fill(i);			/* Fill up last line	*/
#if DEBUG>1
    	    printf("....Returning from comment, lines=%d\n", lines);
#endif
    return(lines);
}

/*==================================================================
		FITS Utilities
 *==================================================================*/
int fit_setmode(newmode) 
/*++++++++++++++++
.PURPOSE  Define the current mode
.RETURNS  The previous mode (0=newlines, 1=block to 2880)
.REMARKS  Should not normally be changed after a file opnening.
-----------------*/
  int newmode;	/* IN: 0=insert newlines, 1=block to 2880 */
{
  int  oldmode;
    oldmode = toFITSd;
    toFITSd = newmode;
    return(oldmode);
}

int fit_getmode() 
/*++++++++++++++++
.PURPOSE  Get the current mode
.RETURNS  The previous mode (0=newlines, 1=block to 2880)
-----------------*/
{
    return (toFITSd);
}

int fit_flushnl(fout, pos)
/*++++++++++++++++
.PURPOSE  Complete the current block with blanks and a newline
.RETURNS  Number of spaces added.
-----------------*/
  FILE *fout;	/* IN: File opened for output	*/
  long pos;	/* IN: Position in the file; -1 for unknown */
{
  int4 i, n; long *apos;
#if VERBOP
  fprintf(stderr, "....fit_flush(%d)\n", pos);
#endif

    apos = fileno(fout) < ITEMS(fpos) ? &fpos[fileno(fout)] : (long *)0;
    if ((pos < 0)  && apos) 	pos = *apos; 	/* Get current position */
    if (pos < 0) 				/* Get current position */
	pos = ftell(fout);
    if (pos < 0) { 
	perror("****fit_flush"); 
	return(-1);
    }
    n = pos%2880;			/* Number of blanks to insert */
    if (n) n = 2880-n;
    if (toFITSd) {
	if ((i = n)) {
	    while (--i > 0) fputc(' ', fout);
	    fputc('\n', fout) ;
	    pos += n ;
	}
    }
    else fprintf(fout, "\n");
    if (ferror(fout)) perror("****fit_flush");
    if (apos) *apos = 0;
#if VERBOP
  fprintf(stderr, "....return from fit_flush: %d\n", i);
#endif
    fflush(fout);
    return(n);
}

int fit_flush(fout, pos)
/*++++++++++++++++
.PURPOSE  Complete the current block with blanks.
.RETURNS  Number of blanks added.
-----------------*/
  FILE *fout;	/* IN: File opened for output	*/
  long pos;	/* IN: Position in the file; -1 for unknown  */
{
  int n; long *apos;
#if VERBOP
  fprintf(stderr, "....fit_flush(%d)\n", pos);
#endif

    apos = fileno(fout) < ITEMS(fpos) ? &fpos[fileno(fout)] : (long *)0;
    if ((pos < 0)  && apos) 	pos = *apos; 	/* Get current position */
    if (pos < 0) 				/* Get current position */
	pos = ftell(fout);
    if (pos < 0) { 
	perror("****fit_flush"); 
	return(-1);
    }
    n = pos%2880;			/* Number of blanks to insert */
    if (n) n = 2880-n;

#if VERBOP
  fprintf(stderr, "....now pos=%d\n", pos);
#endif

    if (toFITSd) {
	while (pos%2880) fputc(' ', fout), pos++;
    }
    else fprintf(fout, "\n");
    if (ferror(fout)) perror("****fit_flush");
    if (apos) *apos = 0;
#if VERBOP
  fprintf(stderr, "....return from fit_flush: %d\n", i);
#endif
    fflush(fout);
    return(n);
}

int fit_prka(fout, kw, val, com)
/*++++++++++++++++
.PURPOSE  Issue a CHARACTER keyword
.RETURNS  Number of lines
.REMARKS  No '=' sign when kw is NULL
-----------------*/
  FILE *fout;	/* IN: File opened for output	*/
  char *kw;	/* IN: Keywordname		*/
  char *val;	/* IN: Keyword value		*/
  char *com;	/* IN: Comment to append	*/
{
  int i, lines; char *p;

    f = fout; if (!com) com = "";
    lines = 0;
	/* Write out the keyword name (limited to 8 bytes) */
    i = prkw(kw, kw ? 1 : 0); 
    fputc('\'', f), i++;
    	/* Write the value (at least 8 bytes within ' */
    for (p=val; *p; p++, i++) {			/* Write out value */
	if (*p == '\n') ;
	else {
	    if (i < 77) goto Write_char ;
	    if ((i == 77) && (*p != '\'')) goto Write_char ;
	}
						/* Write CONTINUE  */
	if (i < 79) fputc('&' , f), i++; 
	if (i < 80) fputc('\'', f), i++; 
	lines +=  comment(i, "");
	i = prkw("CONTINUE", 0);
	fputc(' ', f), i++;
	fputc('\'', f);
	if (*p == '\n') { while (isspace(p[1])) p++; continue; }
	else i++;

      Write_char:
    	if ((*p == '\'') && (i < 79)) fputc('\'', f), i++;
    	fputc(*p, f);
    }
    while (i < 19) fputc(' ', f), i++;	/* At leat 8 bytes */
    if (i < 80) fputc('\'', f), i++;
    lines += comment(i, com);
#if VERBOP
  fprintf(stderr, "\n....fit_prka(%s): fpos is now %d (%d lines)\n", 
	kw, fpos[fileno(f)], lines);
#endif
    return(lines);
}

int fit_prkn(fout, kw, val, com)
/*++++++++++++++++
.PURPOSE  Issue a NUMERIC (right-aligned) keyword
.RETURNS  Number of lines
-----------------*/
  FILE *fout;	/* IN: File opened for output	*/
  char *kw;	/* IN: Keywordname		*/
  char *val;	/* IN: Keyword value		*/
  char *com;	/* IN: Comment to append	*/
{
  int i, len, r; char *p ;

    f = fout; if (!com) com = "";
	/* Write out the keyword name (limited to 8 bytes) */
    i = prkw(kw, 1);
	/* Write out the value right-adjusted */
    for (p=val;  isspace(*p); p++) ;	/* Skip blanks */
    len = match_num(p); 		/* Keep only valid number */
    if (len == 0) len = strlen(p);
#if 0	/* This causes problems with gcc ... */
    e = p + len; x = *e; *e = 0;
    if (len <= 20) fprintf(f, "%20s", p), i += 20;
    else for (; (*p) && (i<80); p++, i++) fputc(*p, f);
    *e = x;
#else
    for (r = 20-len; --r >= 0; i++) fputc(' ', f) ;
    for (; (*p) && (i<80); p++, i++) fputc(*p, f);
#endif
    r = comment(i, com);
#if VERBOP
  fprintf(stderr, "\n....fit_prkn(%s): fpos is now %d \n", 
	kw, fpos[fileno(f)]);
#endif
    return(r);
}

int fit_prki(fout, kw, val, com)
/*++++++++++++++++
.PURPOSE  Issue an INTEGER keyword
.RETURNS  Number of lines
-----------------*/
  FILE *fout;	/* IN: File opened for output	*/
  char *kw;	/* IN: Keywordname		*/
  int  val;	/* IN: Keyword value		*/
  char *com;	/* IN: Comment to append	*/
{
  int i, r;

    f = fout; if (!com) com = "";
	/* Write out the keyword name (limited to 8 bytes) */
    i = prkw(kw, 1);
    	/* Write the value */
    fprintf(f, "%20d", val), i += 20;
    r = comment(i, com);
#if VERBOP
  fprintf(stderr, "\n....fit_prki(%s): fpos is now %d \n", 
	kw, fpos[fileno(f)]);
#endif
    return(r);
}

int fit_prcom(fout, kw, com)
/*++++++++++++++++
.PURPOSE  Issue COMMENT
.RETURNS  Number of lines
-----------------*/
  FILE *fout;	/* IN: File opened for output	*/
  char *kw;	/* IN: Keyword (normally COMMENT or blank) */
  char *com;	/* IN: Comment to append	*/
{
  int i; char old_noslash;

#if DEBUG>1
    printf("prcom(%s): %s", kw, com); inbyte = getchar();
#endif
    old_noslash = noslash; noslash=1;
    f = fout; if (!com) com = "";
	/* Write out the keyword name (limited to 8 bytes) */
    i = prkw(kw, 0);
    i = comment(i, com);
    noslash = old_noslash;
#if VERBOP
  fprintf(stderr, "\n....fit_prcom(%s): fpos is now %d \n", 
	kw, fpos[fileno(f)]);
#endif
    return(i);
}

int fit_prt(fout, kw, com, len)
/*++++++++++++++++
.PURPOSE  Issue COMMENT which may contain newlines; words are not cut.
.RETURNS  Number of lines
-----------------*/
  FILE *fout;	/* IN: File opened for output	*/
  char *kw;	/* IN: Keyword (normally COMMENT or blank) */
  char *com;	/* IN: Comment to append	*/
  int  len;	/* IN: Length of com		*/
{
  int i, lines, bytes, indent; char *p, *e, *n;

#if DEBUG
    printf("prt(%s): %d bytes", kw, len); 
    fwrite(com, 1, len, stdout); inbyte = getchar();
#endif
    f = fout; 
    	/* Remove the trailing spaces */
    len = strnline(com, len);
    e = com + len;
#if DEBUG
    printf("....Trimmed length=%d", len);
#endif
    	/* Count the indentation */
    for (p=com, indent=20; p<e; p += i+1) {
    	i = strloc(p, '\n'); 
    	if (i == 0) continue;
    	for (n=p; *n == ' '; n++) ;
    	bytes = n-p;
    	if (bytes >= indent) continue;
    	indent = bytes;
    }
#if DEBUG
    printf(", indent length=%d\n", indent);
#endif
	/* Write out the keyword name (limited to 8 bytes) */
    bytes = prkw(kw, 0);
    for (p=com, lines=0; p<e; p = n+1, lines++) {
    	if (p != com) bytes = prkw("", 0);	/* Continuation */
    	i = strloc(p, '\n'); n = p+i;
#if DEBUG>1
 	printf("....line of %d bytes reduced to %d bytes\n", i, 
 		strnline(p, i));
#endif
    	i = strnline(p, i);	/* Forget trailing spaces */
    	if ((*p == ' ') && (i>indent)) p += indent, i -= indent;
    	if ((bytes + i) > 80) {	/* Can't write in a single line */
    	    i = 80-bytes;
    	    if (!match_rule(p)) {	/* Find out word separation */
    		while (--i > 0) { if (isspace(p[i])) break; }
    		if (i == 0) { i = 80-bytes; n = p+i-1; }
    		else n = p+i;
    	    }
    	}
    	fwrite(p, 1, i, f);
    	bytes += i; fill(bytes);
    }
    if (lines == 0) fill(bytes), lines=1;
    return(lines);
}

#ifndef XML2FITS
int fit_copy(fout, file)
/*++++++++++++++++
.PURPOSE  Copy a file
.RETURNS  Number of buffers
.REMARKS  filename is assumed to be a *.fits file
-----------------*/
  FILE *fout;	/* IN: File opened for output	*/
  char *file;	/* IN: File to copy		*/
{
  long bytes;
  int len; char *buf; FILE *ifile;

    ifile = fzopen(file);
    if (!ifile) return(0);

    bytes = 0; buf = malloc(BUFb);
    while ((len = fread(buf, 1, BUFb, ifile)) > 0) {
#if DEBUG
	printf("....Read %d bytes (%d asked)\n", len, BUFb);
#endif
    	bytes += fwrite(buf, 1, len, fout);
    	if (ferror(ifile) || ferror(fout)) break;
    }
    free(buf);
    return(bytes/2880);
}

int fit_prfile(fout, file)
/*++++++++++++++++
.PURPOSE  Write out the contents of filename 
.RETURNS  Number of lines
.REMARKS  file is assumed to be a *.fih file
-----------------*/
  FILE *fout;	/* IN: File opened for output	*/
  char *file;	/* IN: File to copy (*.fih)	*/
{
  int i, len, lines; char buf[84]; FILE *ifile;
  int4 atoi() ;

    ifile = fopen(file, RMODE);
    if (!ifile) { perror(file); return(0); }
    f = fout; lines = 0;
    for (i=0; i<ITEMS(NAXIS); i++) NAXIS[i] = 0;
    while (fgets(buf, sizeof(buf), ifile)) {
    	len = strline(buf);
    	fwrite(buf, 1, len, f);		
    	lines += fill(len);
    	if (strmatch("NAXIS*", buf)) {	/* Keep values */
    	    i = atoi(buf+5);
    	    if (i < 0) continue;
    	    if (i >= ITEMS(NAXIS)) continue;
    	    NAXIS[i] = atoi(buf+10);
    	}
    }
    if (ferror(ifile)) perror(file);
    fclose(ifile);
    return(lines);
}
#endif

int fit_naxis(n)
/*++++++++++++++++
.PURPOSE  Retrieve the NAXIS parameter from previously read file.
.RETURNS  Its value
-----------------*/
  int n;	/* IN: The NAXIS to read (1=reclen, 2=records) */
{
    if ((n >= 0) && (n < ITEMS(NAXIS))) return (NAXIS[n]);
    return (0);
}

#ifdef TEST
/*==================================================================
		Test Program
 *==================================================================*/
main() 
/*++++++++++++++++
.PURPOSE  Just a TEST program
-----------------*/
{
  static char buf[512]; FILE *f; int i; char *kw, *val, *com;

    f = stdout; fit_setmode(0);
    while(1) {
        kw = buf;
    	printf("Keyword name: "); if (!gets(kw)) break;
    	val = kw + strlen(kw) + 1;
    	printf("Keyword value: "); if (!gets(val)) break;
    	com = val + strlen(val) + 1;
    	printf("Comment: "); if (!gets(com)) break;
    	if (isdigit(*val)) fit_prkn(f, kw, val, com);
    	else fit_prka(f, kw, val, com);
    	fit_prcom(f, "COMMENT", com);
    }
    fit_flush(stdout, -1L);
}
#endif
